
-- This is an example Postgres sql file that creates some tables
-- for use by Autograph.

create schema financial;
create schema general;
create schema discussion;

create table general.people
    (
    id		serial,
    name	text,
    creation_date timestamp not null default CURRENT_TIMESTAMP,
    primary key(id)
    );

create table general.emails
    (
    id		serial,
    email	text not null,
    primary key(id)
    );

create table general.email_holder
    (
    id		serial,
    person_id	integer not null references general.people(id),
    email_id	integer not null references general.emails(id),
    valid_from	timestamp not null default CURRENT_TIMESTAMP,
    valid_to	timestamp,
    primary key(id)
    );

-- The currencies table is purposely not being used by anything to illustrate
-- how this looks in the diagram.
create table financial.currencies
    (
    id			serial,
    name		text not null,
    code		char(3),
    primary key(id)
    );

create table financial.accounts
    (
    id			serial,
    holder		integer not null references general.people(id),
    creation_date	timestamp not null default CURRENT_TIMESTAMP,
    primary key(id)
    );

create table financial.transactions
    (
    id			serial,
    creation_date	timestamp not null default CURRENT_TIMESTAMP,
    primary key(id)
    );

create table financial.trans_parts
    (
    id			serial,
    transaction_id	integer not null references financial.transactions(id),
    from_acct		integer not null references financial.accounts(id),
    to_acct		integer not null references financial.accounts(id),
    amount		decimal(15,2) not null default 0,
    primary key(id)
    );

create index from_accts_idx on financial.trans_parts(from_acct);
create index to_accts_idx on financial.trans_parts(to_acct);

create table discussion.comments
    (
    id		serial,
    reply_to	integer references discussion.comments(id),
    commentor	integer not null references general.people(id),
    comm_txt	text not null,
    creation_date	timestamp not null default CURRENT_TIMESTAMP,
    primary key(id)
    );

